@@ -24,6 +24,8 @@ type protocol =
   | SSLv23
   | SSLv3
   | TLSv1
+  | TLSv1_1
+  | TLSv1_2
 
 type context
 
@@ -80,6 +82,8 @@ exception Method_error
 exception Context_error
 exception Certificate_error
 exception Cipher_error
+exception Diffie_hellman_error
+exception Ec_curve_error
 exception Private_key_error
 exception Unmatching_keys
 exception Invalid_socket
@@ -95,6 +99,8 @@ let () =
   Callback.register_exception "ssl_exn_context_error" Context_error;
   Callback.register_exception "ssl_exn_certificate_error" Certificate_error;
   Callback.register_exception "ssl_exn_cipher_error" Cipher_error;
+  Callback.register_exception "ssl_exn_diffie_hellman_error" Diffie_hellman_error;
+  Callback.register_exception "ssl_exn_ec_curve_error" Ec_curve_error;
   Callback.register_exception "ssl_exn_private_key_error" Private_key_error;
   Callback.register_exception "ssl_exn_unmatching_keys" Unmatching_keys;
   Callback.register_exception "ssl_exn_invalid_socket" Invalid_socket;
@@ -135,6 +141,10 @@ external embed_socket : Unix.file_descr -> context -> socket = "ocaml_ssl_embed_
 
 external set_cipher_list : context -> string -> unit = "ocaml_ssl_ctx_set_cipher_list"
 
+external init_dh_from_file : context -> string -> unit = "ocaml_ssl_ctx_init_dh_from_file"
+
+external init_ec_from_named_curve : context -> string -> unit = "ocaml_ssl_ctx_init_ec_from_named_curve"
+
 external load_verify_locations : context -> string -> string -> unit = "ocaml_ssl_ctx_load_verify_locations"
 
 external get_verify_result : socket -> int = "ocaml_ssl_get_verify_result"
@@ -180,6 +190,8 @@ external get_subject : certificate -> string = "ocaml_ssl_get_subject"
 
 external file_descr_of_socket : socket -> Unix.file_descr = "ocaml_ssl_get_file_descr"
 
+external set_client_SNI_hostname : socket -> string -> unit = "ocaml_ssl_set_client_SNI_hostname"
+
 external connect : socket -> unit = "ocaml_ssl_connect"
 
 external verify : socket -> unit = "ocaml_ssl_verify"
